import numpy as np
from scipy.sparse import diags
from scipy.linalg import eigh

def C_matrix(N,Cjj,Cg,Cg_big,Cin,Cout):
        #TODO: #5 Add documentation and adapt the introduction of Cg_big
        matrix_diagonals = [(-Cjj)*np.ones(N),(2*Cjj+Cg)*np.ones(N+1),(-Cjj)*np.ones(N)]
        matrix = diags(matrix_diagonals,offsets=[-1,0,1]).toarray()
        matrix[0,0] = Cjj + Cin
        matrix[-1,-1] = Cjj + Cout
        matrix[77,77] = 2*Cjj+ Cg_big
        matrix[-78,-78] = 2*Cjj+ Cg_big
        return matrix

def L_inv_matrix(N,Ljj):
    matrix_diagonals = [(-1/Ljj)*np.ones(N),(2/Ljj)*np.ones(N+1),(-1/Ljj)*np.ones(N)]
    matrix = diags(matrix_diagonals,[-1,0,1]).toarray()
    matrix[0,0] = 1/Ljj
    matrix[-1,-1] = 1/Ljj
    return matrix

def jja_resonances(params):
    Ljj = params[0]*1e-9
    Cjj = params[1]*1e-15
    Cg = params[2]*1e-18
    Cg_big = params[3]*1e-15
    Cin = params[4]*1e-15
    
    N = 170
    Cout = 0

    # RESOLVER EL PROBLEMA DE REDONDEO DE NUMEROS PEQUEÑOS.

    # Eigensolution of the linearized circuit matrix for a Josephson junction chain with the following parameters:
    # N: number of junctions
    # Cjj: junction capacitance
    # Ljj: junction inductance
    # Cg: ground capacitance
    # Cin: input capacitance
    # Cout: output capacitance

    # Return the frequency in Hz.

    # References:
    #   https://theses.hal.science/tel-01369020
    #   https://journals.aps.org/prb/abstract/10.1103/PhysRevB.92.104508

    # Solving C^-1/2 L^-1 C^-1/2 phi = \omega^2 phi

    eigenvalues_C, eigenvectors_C = eigh(C_matrix(N,Cjj,Cg,Cg_big,Cin,Cout))
    Lambda_inv_sqrt = np.diag(1 / np.sqrt(eigenvalues_C))
    C_inv_sqrt = np.dot(eigenvectors_C, np.dot(Lambda_inv_sqrt, eigenvectors_C.T)) # spectral decomposition of C^-1/2
    matrix_operation = np.dot(C_inv_sqrt, np.dot(L_inv_matrix(N,Ljj), C_inv_sqrt)) # C^-1/2 L^-1 C^-1/2
    eigvals = eigh(matrix_operation, eigvals_only=True) # eigenvalues and eigvecs of C^-1/2 L^-1 C^-1/2

    return np.sqrt(eigvals)/2/np.pi

def jja_eigensys(params, **kwargs):
    Ljj = params[0]*1e-9
    Cjj = params[1]*1e-15
    Cg = params[2]*1e-18
    Cg_big = params[3]*1e-15
    Cin = params[4]*1e-15
    
    N = 170
    Cout = 0

    eigenvalues_C, eigenvectors_C = eigh(C_matrix(N,Cjj,Cg,Cg_big,Cin,Cout))
    Lambda_inv_sqrt = np.diag(1 / np.sqrt(eigenvalues_C))
    C_inv_sqrt = np.dot(eigenvectors_C, np.dot(Lambda_inv_sqrt, eigenvectors_C.T)) # spectral decomposition of C^-1/2
    matrix_operation = np.dot(C_inv_sqrt, np.dot(L_inv_matrix(N,Ljj), C_inv_sqrt)) # C^-1/2 L^-1 C^-1/2
    eigvals, eigvecs = eigh(matrix_operation, **kwargs) # eigenvalues and eigvecs of C^-1/2 L^-1 C^-1/2

    return np.sqrt(eigvals), eigvecs # Returns \omega_k, \Phi_k








import matplotlib.pyplot as plt
import numpy as np
from .qubit_base import QubitBase
from typing import Any, Dict, Optional, Tuple, Union, Iterable
from .operators import sigma_z, sigma_y, sigma_x
from .utilities import sin_kphi_operator, cos_kphi_operator

class Andreev(QubitBase):
    PARAM_LABELS = {
        'Ec': r'$E_C$',
        'Gamma': r'$\Gamma$',
        'delta_Gamma': r'$\delta \Gamma$',
        'er': r'$\epsilon_r$',
        'phase': r'$\Phi_{ext} / \Phi_0$',
        'ng': r'$n_g$'
    }
    
    OPERATOR_LABELS = {
    'n_operator': r'\hat{n}',
    'd_hamiltonian_d_ng': r'\partial \hat{H} / \partial n_g',
    'd_hamiltonian_d_deltaGamma': r'\partial \hat{H} / \partial \delta \Gamma',
    'd_hamiltonian_d_er': r'\partial \hat{H} / \partial \epsilon_r',
    }
    
    def __init__(self, Ec, Gamma, delta_Gamma, er, phase, ng, n_cut, Delta = 40):
        """
        Initializes the Ferbo class with the given parameters.

        Parameters
        ----------
        Ec : float
            Charging energy.
        Gamma : float
            Coupling strength.
        delta_Gamma : float
            Coupling strength difference.
        er : float
            Energy relaxation rate.
        phase : float
            External magnetic phase.
        dimension : int
            Dimension of the Hilbert space.
        ng : float
            Charge offset.
        n_cut : int
            Maximum number of charge states.
        Delta : float
            Superconducting gap.
        """
        
        self.Ec = Ec
        self.Gamma = Gamma
        self.delta_Gamma = delta_Gamma
        self.er = er
        self.phase = phase
        self.ng = ng
        self.n_cut = n_cut
        self.dimension = 2 * (self.n_cut * 4 + 1)
        self.Delta = Delta
        
        super().__init__(self.dimension)
        
    
    def n_operator(self) -> np.ndarray:
        """
        Returns the charge number operator adjusted for half-charge translations.

        Returns
        -------
        np.ndarray
            The charge number operator.
        """
        n_values = np.arange(-self.n_cut, self.n_cut+1/2 , 1/2) - self.ng * np.ones(self.dimension // 2)
        n_matrix = np.diag(n_values)
        return np.kron(np.eye(2), n_matrix)  
    
    def jrl_potential(self) -> np.ndarray:
        """
        Returns the Josephson Resonance Level potential in the half-charge basis.

        Returns
        -------
        np.ndarray
            The Josephson Resonance Level potential.
        """
        
        Gamma_term = -self.Gamma * np.kron(sigma_z(), cos_kphi_operator(1, self.dimension // 2, self.phase/2))
        delta_Gamma_term = - self.delta_Gamma * np.kron(sigma_y(), sin_kphi_operator(1, self.dimension // 2, self.phase/2))
        e_r_term = self.er * np.kron(sigma_x(), np.eye(self.dimension // 2))
        
        return Gamma_term + delta_Gamma_term + e_r_term
            
    # def zazunov_potential(self) -> np.ndarray:
        
    def hamiltonian(self) -> np.ndarray:
        """
        Returns the Hamiltonian of the system.

        Returns
        -------
        np.ndarray
            The Hamiltonian of the system.
        """
        n_x = self.delta_Gamma/4/(self.Gamma+self.Delta)
        n_op = self.n_operator() + n_x * np.kron(sigma_x(),np.eye(self.dimension//2))
        
        charge_term = 4 * self.Ec * n_op @ n_op

        potential = self.jrl_potential()
        return charge_term + potential
    
    def d_hamiltonian_d_ng(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the number of charge offset.
        
        Returns
        -------
        np.ndarray
            The derivative of the Hamiltonian with respect to the number of charge offset.
        
        """
        return  - 8 * self.Ec * self.n_operator()
    
    def d_hamiltonian_d_phase(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the external magnetic phase.

        Returns
        -------
        np.ndarray
            The derivative of the Hamiltonian with respect to the external magnetic phase.
        """
        return NotImplementedError("Not implemented yet")
                
    def d_hamiltonian_d_er(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the energy relaxation rate.

        Returns
        -------
        Qobj
            The derivative of the Hamiltonian with respect to the energy relaxation rate.
        """
        # return - np.kron(np.eye(self.dimension // 2),sigma_z())
        return NotImplementedError("Not implemented yet")
    
    def d_hamiltonian_d_deltaGamma(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the coupling strength difference.

        Returns
        -------
        Qobj
            The derivative of the Hamiltonian with respect to the coupling strength difference.
        """
        return NotImplementedError("Not implemented yet")
        # if self.flux_grouping == 'L':
        #     phase_op = self.phase_operator()[::2,::2]
        # else:
        #     phase_op = self.phase_operator()[::2,::2] - self.phase * np.eye(self.dimension // 2)
        # return - np.kron(sinm(phase_op/2),sigma_y())
    
    def wavefunction(
        self, 
        which: int = 0,
        phi_grid: np.ndarray = None,
        esys: Tuple[np.ndarray, np.ndarray] = None,
        basis: str = 'default'
        ) -> Dict[str, Any]:
        """
        Returns a wave function in the phi basis.

        Parameters
        ----------
        which : int, optional
            Index of desired wave function (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.
        basis : str, optional
            Basis in which to return the wave function ('default' or 'abs') (default is 'default').

        Returns
        -------
        Dict[str, Any]
            Wave function data containing basis labels, amplitudes, and energy.
        """
        return NotImplementedError("Not implemented yet")
        # if esys is None:
        #     evals_count = max(which + 1, 3)
        #     evals, evecs = self.eigensys(evals_count)
        # else:
        #     evals, evecs = esys
            
        # dim = self.dimension//2
        
        # if basis == 'default':
        #     evecs = evecs.T
        # elif basis == 'abs':
        #     U = (1 / np.sqrt(2)) * np.array([[1, 1], [1, -1]])
        #     change_of_basis_operator = np.kron(np.eye(dim), U)
        #     evecs = (change_of_basis_operator @ evecs).T        
                        
        # if phi_grid is None:
        #     phi_grid = np.linspace(-5 * np.pi, 5 * np.pi, 151)

        # phi_basis_labels = phi_grid
        # wavefunc_osc_basis_amplitudes = evecs[which, :]
        # phi_wavefunc_amplitudes = np.zeros((2, len(phi_grid)), dtype=np.complex_)
        # phi_osc = self.phi_osc()
        # for n in range(dim):
        #     phi_wavefunc_amplitudes[0] += wavefunc_osc_basis_amplitudes[2 * n] * self.harm_osc_wavefunction(n, phi_basis_labels, phi_osc)
        #     phi_wavefunc_amplitudes[1] += wavefunc_osc_basis_amplitudes[2 * n + 1] * self.harm_osc_wavefunction(n, phi_basis_labels, phi_osc)

        # return {
        #     "basis_labels": phi_basis_labels,
        #     "amplitudes": phi_wavefunc_amplitudes,
        #     "energy": evals[which]
        # }
        
    def potential(self, phi: Union[float, np.ndarray]) -> np.ndarray:
        """
        Calculates the potential energy for given values of phi.

        Parameters
        ----------
        phi : Union[float, np.ndarray]
            The phase values at which to calculate the potential.

        Returns
        -------
        np.ndarray
            The potential energy values.
        """
        return NotImplementedError("Not implemented yet")
        # phi_array = np.atleast_1d(phi)
        # evals_array = np.zeros((len(phi_array), 2))
        # phi_ext = 2 * np.pi * self.phase

        # for i, phi_val in enumerate(phi_array):
        #     if self.flux_grouping == 'ABS':
        #         inductive_term = 0.5 * self.El * phi_val**2 * np.eye(2)
        #         andreev_term = -self.Gamma * np.cos((phi_val + self.phase) / 2) * sigma_x() - self.delta_Gamma * np.sin((phi_val + self.phase) / 2) * sigma_y() - self.er * sigma_z()
        #     elif self.flux_grouping == 'L':
        #         inductive_term = 0.5 * self.El * (phi_val + phi_ext)**2 * np.eye(2)
        #         andreev_term = -self.Gamma * np.cos(phi_val / 2) * sigma_x() - self.delta_Gamma * np.sin(phi_val / 2) * sigma_y() - self.er * sigma_z()
            
        #     potential_operator = inductive_term + andreev_term
        #     evals_array[i] = eigh(
        #         potential_operator,
        #         eigvals_only=True,
        #         check_finite=False,
        # )

        # return evals_array
    
    def tphi_1_over_f(
        self, 
        A_noise: float, 
        i: int, 
        j: int, 
        noise_op: str,
        esys: Tuple[np.ndarray, np.ndarray] = None,
        get_rate: bool = False,
        **kwargs
        ) -> float:
        """
        Calculates the 1/f dephasing time (or rate) due to an arbitrary noise source.

        Parameters
        ----------
        A_noise : float
            Noise strength.
        i : int
            State index that along with j defines a qubit.
        j : int
            State index that along with i defines a qubit.
        noise_op : str
            Name of the noise operator, typically Hamiltonian derivative w.r.t. noisy parameter.
        esys : Tuple[np.ndarray, np.ndarray], optional
            Precomputed eigenvalues and eigenvectors (default is None).
        get_rate : bool, optional
            Whether to return the rate instead of the Tphi time (default is False).

        Returns
        -------
        float
            The 1/f dephasing time (or rate).
        """
        p = {"omega_ir": 2 * np.pi * 1, "omega_uv": 3 * 2 * np.pi * 1e6, "t_exp": 10e-6}
        p.update(kwargs)
                
        if esys is None:
            evals, evecs = self.eigensys(evals_count=max(j, i) + 1)
        else:
            evals, evecs = esys

        noise_operator = getattr(self, noise_op)()    
        dEij_d_lambda = np.abs(evecs[i].conj().T @ noise_operator @ evecs[i] - evecs[j].conj().T @ noise_operator @ evecs[j])

        rate = (dEij_d_lambda * A_noise * np.sqrt(2 * np.abs(np.log(p["omega_ir"] * p["t_exp"]))))
        rate *= 2 * np.pi * 1e9 # Convert to rad/s

        return rate if get_rate else 1 / rate
    
    def tphi_1_over_f_flux(
        self, 
        A_noise: float = 1e-6,
        i: int = 0, 
        j: int = 1, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        get_rate: bool = False, 
        **kwargs
        ) -> float:
        return self.tphi_1_over_f(A_noise, i, j, 'd_hamiltonian_d_phase', esys=esys, get_rate=get_rate, **kwargs)

    def plot_wavefunction(
        self, 
        which: Union[int, Iterable[int]] = 0, 
        phi_grid: np.ndarray = None, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        scaling: Optional[float] = None,
        basis: str = 'default',
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the wave function in the phi basis.

        Parameters
        ----------
        which : Union[int, Iterable[int]], optional
            Index or indices of desired wave function(s) (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.
        esys : Tuple[np.ndarray, np.ndarray], optional
            Precomputed eigenvalues and eigenvectors.
        basis: str, optional
            Basis in which to return the wavefunction ('default' or 'abs') (default is 'default').
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if isinstance(which, int):
            which = [which]
            
        potential = self.potential(phi=phi_grid)

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle())
        else:
            fig, ax = fig_ax
        
        ax.plot(phi_grid/2/np.pi, potential[:, 0], color='black', label='Potential')
        ax.plot(phi_grid/2/np.pi, potential[:, 1], color='black')

        for idx in which:
            wavefunc_data = self.wavefunction(which=idx, phi_grid=phi_grid, esys=esys, basis = basis)
            phi_basis_labels = wavefunc_data["basis_labels"]
            wavefunc_amplitudes = wavefunc_data["amplitudes"]
            wavefunc_energy = wavefunc_data["energy"]

            ax.plot(
                phi_basis_labels/2/np.pi,
                wavefunc_energy + scaling * (wavefunc_amplitudes[0].real + wavefunc_amplitudes[0].imag),
                # color="blue",
                label=rf"$\Psi_{idx} \uparrow $"
                )
            ax.plot(
                phi_basis_labels/2/np.pi, 
                wavefunc_energy + scaling * (wavefunc_amplitudes[1].real + wavefunc_amplitudes[1].imag),
                # color="red",
                label=rf"$\Psi_{idx} \downarrow $"
                )

        ax.set_xlabel(r"$\Phi / \Phi_0$")
        ax.set_ylabel(r"$\psi(\varphi)$, Energy [GHz]")
        ax.legend()
        ax.grid(True)

        return fig, ax      


import numpy as np
import matplotlib.pyplot as plt
import scqubits.utils.plotting as plot
from abc import ABC, abstractmethod
from scipy.linalg import expm, eigh
from scipy.special import factorial, pbdv
from .storage import SpectrumData
from tqdm.notebook import tqdm
from scipy.constants import hbar, k, h, e
from scipy.special import k0
from typing import Tuple, List, Union, Any, Callable, Optional

class QubitBase(ABC):
    PARAM_LABELS = {}
    OPERATOR_LABELS = {}
    
    def __init__(self, dimension:int):
        self.dimension = dimension
    
    def __repr__(self) -> str:
        """
        Returns a string representation of the QubitBase instance.

        Returns
        -------
        str
            A string representation of the QubitBase instance.
        """
        init_params = [param for param in self.__dict__ if not param.startswith('_')]
        init_dict = {name: getattr(self, name) for name in init_params}
        return f"{type(self).__name__}(**{init_dict!r})"
    
    @abstractmethod
    def n_operator(self) -> np.ndarray:
        """
        Returns the number operator for the qubit.

        Returns
        -------
        ndarray
            The number operator for the qubit.
        """
        pass
    
    def displacement_operator(self) -> np.ndarray:
        """
        Returns the displacement operator.

        Returns
        -------
        np.ndarray
            The displacement operator.
        """
        return expm(-1j*2*np.pi*self.n_operator())
    
    @abstractmethod
    def hamiltonian(self) -> np.ndarray:
        """
        Returns the Hamiltonian for the qubit.

        Returns
        -------
        ndarray
            The Hamiltonian for the qubit.
        """
        pass
    
    def eigensys(self, evals_count: int = 6) -> Tuple[np.ndarray, np.ndarray]:
        """
        Calculates eigenvalues and corresponding eigenvectors using scipy.linalg.eigh.

        Parameters
        ----------
        evals_count : int, optional
            Number of desired eigenvalues/eigenstates (default is 6).

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Eigenvalues and eigenvectors as numpy arrays.
        """
        hamiltonian_mat = self.hamiltonian()
        evals, evecs = eigh(
            hamiltonian_mat,
            eigvals_only=False,
            subset_by_index=(0, evals_count - 1),
            # check_finite=False,
        )
        return evals, evecs
    
    def eigenvals(self, evals_count: int = 6) -> np.ndarray:
        """
        Calculates eigenvalues using scipy.linalg.eigh.

        Parameters
        ----------
        evals_count : int, optional
            Number of desired eigenvalues (default is 6).

        Returns
        -------
        np.ndarray
            Eigenvalues as a numpy array.
        """
        hamiltonian_mat = self.hamiltonian()
        evals = eigh(
            hamiltonian_mat,
            eigvals_only=True,
            subset_by_index=(0, evals_count - 1),
            check_finite=False,
        )
        return np.sort(evals)
    
    def get_spectrum_vs_paramvals(self, param_name: str, param_vals: List[float], evals_count: int = 6, subtract_ground: bool = False)  -> SpectrumData:
        """
        Calculates the eigenenergies and eigenstates for a range of parameter values.

        Parameters
        ----------
        param_name : str
            The name of the parameter to vary.
        param_vals : List[float]
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        subtract_ground : bool, optional
            Whether to subtract the ground state energy from the eigenenergies (default is False).

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The eigenenergies and eigenstates for the range of parameter values.
        """
        eigenenergies_array = []
        eigenstates_array = []
        
        initial_value = getattr(self, param_name)
        
        for val in tqdm(param_vals, leave=False):
            self.set_param(param_name, val)
            eigenenergies, eigenstates = self.eigensys(evals_count)
            eigenenergies_array.append(eigenenergies)
            eigenstates_array.append(eigenstates)
            
        self.set_param(param_name, initial_value)
            
        spectrum_data = SpectrumData(
            energy_table=np.array(eigenenergies_array),
            system_params=self.__dict__,
            param_name=param_name,
            param_vals=np.array(param_vals),
            state_table=np.array(eigenstates_array)
        )

        if subtract_ground:
            spectrum_data.subtract_ground()
        
        return spectrum_data  
        
    def matrixelement_table(self, operator: str, evecs: np.ndarray = None, evals_count: int = 6) -> np.ndarray:
        """
        Returns a table of matrix elements for a given operator with respect to the eigenstates.

        Parameters
        ----------
        operator : str
            The name of the operator.
        evecs : np.ndarray, optional
            The eigenstates (default is None, in which case they are calculated).
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).

        Returns
        -------
        np.ndarray
            The table of matrix elements.
        """
        
        if evecs is None:
            _, evecs = self.eigensys(evals_count)
            
        operator_matrix = getattr(self, operator)()
        matrix_elements = evecs.conj().T @ operator_matrix @ evecs
        return matrix_elements


    def get_matelements_vs_paramvals(self, operators: Union[str, List[str]], param_name: str, param_vals: np.ndarray, evals_count: int = 6) -> SpectrumData:
        """
        Calculates the matrix elements for a list of operators over a range of parameter values.

        Parameters
        ----------
        operators : Union[str, List[str]]
            The name(s) of the operator(s).
        param_name : str
            The name of the parameter to vary.
        param_vals : np.ndarray
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).

        Returns
        -------
        Dict[str, Dict[str, np.ndarray]]
            The matrix elements for the operators over the range of parameter values.
        """
        if isinstance(operators, str):
            operators = [operators]
                        
        paramvals_count = len(param_vals)
        eigenenergies_array = np.empty((paramvals_count, evals_count))
        eigenstates_array = np.empty((paramvals_count, self.dimension, evals_count), dtype=np.complex128)
        matrixelem_tables = {operator: np.empty((paramvals_count, evals_count, evals_count), dtype=complex) for operator in operators}
        
        initial_value = getattr(self, param_name)
            
        for idx, val in enumerate(tqdm(param_vals, leave=False)):
            self.set_param(param_name, val)
            eigenenergies, eigenstates = self.eigensys(evals_count)
            eigenenergies_array[idx] = eigenenergies
            eigenstates_array[idx] = eigenstates
            
            for operator in operators:
                matrix_elements = self.matrixelement_table(operator, evecs=eigenstates, evals_count=evals_count)
                matrixelem_tables[operator][idx] = matrix_elements
                
        self.set_param(param_name, initial_value)
                            
        spectrum_data = SpectrumData(
            energy_table=eigenenergies_array,
            system_params=self.__dict__,
            param_name=param_name,
            param_vals=param_vals,
            state_table=eigenstates_array,
            matrixelem_table=matrixelem_tables
        )
        
        return spectrum_data
    
    def harm_osc_wavefunction(self, n: int, x: Union[float, np.ndarray], l_osc: float) -> Union[float, np.ndarray]:
        """
        Returns the value of the harmonic oscillator wave function.
        
        The wave function is computed using the parabolic cylinder function Dν(z), 
        which satisfies the Weber differential equation. This implementation 
        is based on the connection between the wave function of the harmonic 
        oscillator and the solutions to the Weber equation.

        Parameters
        ----------
        n : int
            Index of wave function, n=0 is ground state.
        x : Union[float, np.ndarray]
            Coordinate(s) where wave function is evaluated.
        l_osc : float
            Oscillator length.

        Returns
        -------
        Union[float, np.ndarray]
            Value of harmonic oscillator wave function.

        References
        ----------
        - ParabolicCylinderD[ν, z]: https://reference.wolfram.com/language/ref/ParabolicCylinderD.html
        - The wave functions of the quantum harmonic oscillator are proportional 
        to the parabolic cylinder functions Dν(z).

        """
        result = pbdv(n, x / l_osc) / np.sqrt(np.sqrt(2 * np.pi) * factorial(n))
        return result[0]
    
    def plot_matrixelements(
        self,
        operator: str,
        evecs: np.ndarray = None,
        evals_count: int = 6,
        mode: str = "abs",
        show_numbers: bool = False,
        show_colorbar: bool = True,
        show3d: bool = True,
        **kwargs,
    ) -> Union[Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]], Tuple[plt.Figure, plt.Axes]]:
        """
        Plots the matrix elements for a given operator with respect to the eigenstates.

        Parameters
        ----------
        operator : str
            The name of the operator.
        evecs : np.ndarray, optional
            The eigenstates (default is None, in which case they are calculated).
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        mode : str, optional
            The mode for displaying matrix elements ('abs', 'real', 'imag') (default is 'abs').
        show_numbers : bool, optional
            Whether to show the matrix element values as numbers (default is False).
        show_colorbar : bool, optional
            Whether to show a colorbar (default is True).
        show3d : bool, optional
            Whether to show a 3D plot (default is True).

        Returns
        -------
        Union[Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]], Tuple[plt.Figure, plt.Axes]]
            The figure and axes of the plot.
        """
        # Obtener la tabla de elementos de la matriz utilizando el método matrixelement_table
        matrix_elements = self.matrixelement_table(operator, evecs, evals_count)
        
        modefunction = {
            "abs": np.abs,
            "real": np.real,
            "imag": np.imag
        }.get(mode, None)
        
        if modefunction is None:
            raise ValueError(f"Unsupported mode: {mode}")

        matrix_elements = modefunction(matrix_elements)

        if show3d:
            return plot.matrix(
                matrix_elements,
                mode=mode,
                show_numbers=show_numbers,
                **kwargs,
            )

        return plot.matrix2d(
            matrix_elements,
            mode=mode,
            show_numbers=show_numbers,
            show_colorbar=show_colorbar,
            **kwargs,
        )
        
    def t1_capacitive(
        self,
        i: int = 1, 
        j: int = 0, 
        Q_cap: Union[float, Callable] = None,
        T: float = 0.015, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        matrix_elements: np.ndarray = None, 
        get_rate: bool = False,
        noise_op: Optional[np.ndarray] = None
        ) -> float:
        
        if Q_cap is None:
            Q_cap_fun = lambda omega: 1e6 * (2 * np.pi * 6e9 / np.abs(omega))**0.7 # Assuming that Ec is in GHz
        elif callable(Q_cap):
            Q_cap_fun = Q_cap
        else:
            Q_cap_fun = lambda omega: Q_cap

        def spectral_density(omega, T):
            # Assuming that Ec is in GHz
            return 8 * self.Ec / Q_cap_fun(omega) * 1/np.tanh(hbar * np.abs(omega) / (2 * k * T))

        noise_op = noise_op or self.n_operator()
            
        if esys is None:
            evals, evecs = self.eigensys(evals_count=max(i, j) + 1)
        else:
            evals, evecs = esys
            
        omega = 2 * np.pi * (evals[i] - evals[j]) * 1e9  # Convert to rad/s
        
        s = spectral_density(omega, T)
        if matrix_elements is None:
            matrix_elements = self.matrixelement_table('n_operator', evecs=evecs, evals_count=max(i, j) + 1)
        matrix_element = np.abs(matrix_elements[i, j])

        rate = 2 * np.pi * np.abs(matrix_element)**2 * s
        rate *= 1e9
        
        return rate if get_rate else 1 / rate
    
    def t1_inductive(
        self, 
        i: int = 1, 
        j: int = 0,
        Q_ind: float = None,
        T: float = 0.015, 
        esys: Tuple[np.ndarray, np.ndarray] = None,
        matrix_elements: np.ndarray = None, 
        get_rate: bool = False,
        ) -> float:
        
        if Q_ind is None:
            Q_ind_fun = lambda omega: 500e6 * (
                k0(h * 0.5e9 / (2 * k * T)) * np.sinh(h * 0.5e9 / (2 * k * T))
                / (
                    k0(hbar * np.abs(omega) / (2 * k * T))
                    * np.sinh(hbar * np.abs(omega) / (2 * k * T))
                )
            )
        elif callable(Q_ind):
            Q_ind_fun = Q_ind
        else:
            Q_ind_fun = lambda omega: Q_ind
        
        def spectral_density(omega, T):
            return 2 * self.El / Q_ind_fun(omega) * 1 / np.tanh(hbar * np.abs(omega) / (2 * k * T))
            
        if esys is None:
            evals, evecs = self.eigensys(evals_count=max(i, j) + 1)
        else:
            evals, evecs = esys
            
        omega = 2 * np.pi * (evals[i] - evals[j]) * 1e9  # Convert to rad/s
        s = spectral_density(omega, T)
        
        if matrix_elements is None:
            matrix_elements = self.matrixelement_table('phase_operator', evecs=evecs, evals_count=max(i, j) + 1)
        matrix_element = np.abs(matrix_elements[i, j])

        rate = 2 * np.pi * matrix_element**2 * s
        rate *= 1e9  # Convert to GHz
        return rate if get_rate else 1 / rate
    
    def t1_flux_bias_line(
        self, 
        i: int = 1, 
        j: int = 0,
        M: float = 2500,
        Z: float = 50,
        T: float = 0.015, 
        esys: Tuple[np.ndarray, np.ndarray] = None,
        matrix_elements: np.ndarray = None, 
        get_rate: bool = False,
        ) -> float:
        
        def spectral_density(omega, T):
            return 4 * np.pi**2 * M**2 * np.abs(omega) * 1e9 * h / Z * (1 + 1/np.tanh(hbar * np.abs(omega) / (2 * k * T)))
            
        if esys is None:
            evals, evecs = self.eigensys(evals_count=max(i, j) + 1)
        else:
            evals, evecs = esys
            
        omega = 2 * np.pi * (evals[i] - evals[j]) * 1e9  # Convert to rad/s
        s = spectral_density(omega, T)
        
        if matrix_elements is None:
            matrix_elements = self.matrixelement_table('d_hamiltonian_d_phase', evecs=evecs, evals_count=max(i, j) + 1)
        matrix_element = np.abs(matrix_elements[i, j])

        rate = 2 * np.pi * matrix_element**2 * s
        rate *= 1e9  # Convert to GHz
        return rate if get_rate else 1 / rate
        
    def get_t1_vs_paramvals(
        self, 
        noise_channels: Union[str, List[str]],
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> SpectrumData:
        """
        Calculates the T1 times for given noise channels over a range of parameter values.

        Parameters
        ----------
        noise_channels : Union[str, List[str]]
            The noise channels to calculate ('capacitive', 'inductive', etc.).
        param_name : str
            The name of the parameter to vary.
        param_vals : np.ndarray
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments to pass to the T1 calculation method.

        Returns
        -------
        SpectrumData
            The T1 times for the specified noise channels over the range of parameter values.
        """
        if isinstance(noise_channels, str):
            noise_channels = [noise_channels]
            
            if spectrum_data is not None:
                param_name = spectrum_data.param_name
                param_vals = spectrum_data.param_vals
                evals_count = spectrum_data.energy_table.shape[1]
            elif param_name is None or param_vals is None:
                raise ValueError("If spectrum_data is None, param_name and param_vals must be provided.")
                        
        if 'capacitive' in noise_channels:
            spectrum_data = self.get_t1_capacitive_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'inductive' in noise_channels:
            spectrum_data = self.get_t1_inductive_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'charge_impedance' in noise_channels:
            spectrum_data = self.get_t1_charge_impedance_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'critical_current' in noise_channels:
            spectrum_data = self.get_t1_critical_current_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'flux_bias_line' in noise_channels:
            spectrum_data = self.get_t1_flux_bias_line_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'flux_noise' in noise_channels:
            spectrum_data = self.get_t1_1_over_f_flux_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
        if 'Andreev' in noise_channels:
            spectrum_data = self.get_t1_er_vs_paramvals(param_name, param_vals, evals_count, spectrum_data, **kwargs)
            
        return spectrum_data

    def get_t1_capacitive_vs_paramvals(
        self, 
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        evals_count: int = 6, 
        spectrum_data: SpectrumData = None,
        Q_cap: Union[float, Callable] = None,
        T: float = 0.015, 
        **kwargs
        ) -> SpectrumData:
        """
        Calculates the T1 times for capacitive noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        Q_cap : Union[float, Callable], optional
            The capacitance quality factor or a function that returns it (default is None).
        T : float, optional
            The temperature (default is 0.015).
        **kwargs
            Additional arguments to pass to the T1 calculation method.

        Returns
        -------
        SpectrumData
            The T1 times for capacitive noise over the range of parameter values.
        """
        if Q_cap is None:
            Q_cap_fun = lambda omega: 1e6 * (2 * np.pi * 6e9 / np.abs(omega))**0.7 # Assuming that Ec is in GHz
        elif callable(Q_cap):
            Q_cap_fun = Q_cap
        else:
            Q_cap_fun = lambda omega: Q_cap

        def spectral_density(omega, T):
            # Assuming that Ec is in GHz
            return 8 * self.Ec / Q_cap_fun(omega) * 1/np.tanh(hbar * np.abs(omega) / (2 * k * T))

        noise_operator = 'n_operator'
        noise_channel = 'capacitive'
        
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel, T, **kwargs
        )
    
    def get_t1_inductive_vs_paramvals(
        self, 
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        evals_count: int = 6, 
        spectrum_data: SpectrumData = None,
        Q_ind: float = None,
        T: float = 0.015, 
        **kwargs
        ) -> SpectrumData:
        """
        Calculates the T1 times for inductive noise over a range of parameter values.

        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        Q_ind : float, optional
            The inductance quality factor (default is 500e6).
        T : float, optional
            The temperature (default is 0.015).
        **kwargs
            Additional arguments to pass to the T1 calculation method.

        Returns
        -------
        SpectrumData
            The T1 times for inductive noise over the range of parameter values.
        """
        if Q_ind is None:
            Q_ind_fun = lambda omega: 500e6 * (
                k0(h * 0.5e9 / (2 * k * T)) * np.sinh(h * 0.5e9 / (2 * k * T))
                / (
                    k0(hbar * np.abs(omega) / (2 * k * T))
                    * np.sinh(hbar * np.abs(omega) / (2 * k * T))
                )
            )
        elif callable(Q_ind):
            Q_ind_fun = Q_ind
        else:
            Q_ind_fun = lambda omega: Q_ind
        def spectral_density(omega, T):
            return 2 * self.El / Q_ind_fun(omega) * 1 / np.tanh(hbar * np.abs(omega) / (2 * k * T))

        noise_operator = 'phase_operator'
        noise_channel = 'inductive'
        
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel, T, **kwargs
        )
        
    def get_t1_charge_impedance_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        Z: float = 50,
        T: float = 0.015,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the T1 times for charge impedance noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        Z : float, optional
            The impedance (default is 50).
        T : float, optional
            The temperature (default is 0.015).
        **kwargs
        
        Returns
        -------
        SpectrumData
            The T1 times for charge impedance noise over the range of parameter values.
        """
        def spectral_density(omega, T):
            Rk = h / ((2 * e)**2)
            return omega/1e9 / Rk * Z * (1 + 1/np.tanh(hbar * np.abs(omega) / (2 * k * T)))

        noise_operator = 'n_operator'
        noise_channel = 'charge_impedance'
        
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel, T, **kwargs
        )
        
    def get_t1_flux_bias_line_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        M: float = 2500,
        Z: float = 50,
        T: float = 0.015,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the T1 times for flux bias line noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        M : float, optional
            The mutual inductance in units of Phi_0 = h/(2e) (default is 26000).
        Z : float, optional
            The impedance (default is 50).
        T : float, optional
            The temperature (default is 0.015).
        **kwargs
            Additional arguments to pass to the T1 calculation method.
            
        Returns
        -------
        SpectrumData
            The T1 times for flux bias line noise over the range of parameter values.
        """
        
        def spectral_density(omega, T):
            return 4 * np.pi**2 * M**2 * np.abs(omega) * 1e9 * h / Z * (1 + 1/np.tanh(hbar * np.abs(omega) / (2 * k * T)))
        
        noise_operator = 'd_hamiltonian_d_phase'
        noise_channel = 'flux_bias_line'
        
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel, T, **kwargs
        )
        
    def get_t1_1_over_f_flux_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        A_noise: float = 1e-6,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the T1 times for 1/f flux noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        A_noise : float, optional
            The amplitude of the noise (default is 1e-6).
        **kwargs
            Additional arguments to pass to the T1 calculation method.
            
        Returns
        -------
        SpectrumData
            The T1 times for 1/f flux noise over the range of parameter values.
        """
        def spectral_density(omega, T):
            return 2 * np.pi * A_noise**2 / np.abs(omega)
        
        noise_operator = 'd_hamiltonian_d_phase'
        noise_channel = 'flux_noise'
        
        T = 0.015
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel,T, **kwargs
        )

    def get_t1_critical_current_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        A_noise: float = 1e-7,
        N: int = 100,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the T1 times for critical current noise over a range of parameter values.

        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        A_noise : float, optional
            The amplitude of the noise (default is 1e-7).
        N : int, optional
            The number of junctions (default is 100).

        Returns
        -------
        SpectrumData
            The T1 times for critical current noise over the range of parameter values.
        """
        
        def spectral_density(omega, T):
            return 2 * np.pi * (A_noise * self.El / np.sqrt(N))**2 / omega * 1e9
        
        noise_operator = 'd_hamiltonian_d_EL'
        noise_channel = 'critical_current'
        
        T = 0.015
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel,T, **kwargs
        )
        
    def get_t1_er_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        A_noise: float = 0.04,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the T1 times for the Fermi level noise over a range of parameter values.

        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        A_noise : float, optional
            The amplitude of the noise (default is 0.04 GHz).


        Returns
        -------
        SpectrumData
            The T1 times for Fermi level noise over the range of parameter values.
        """
        
        def spectral_density(omega, T):
            return 2 * np.pi * 1e9 * A_noise**2 *np.abs(1/(omega*1e-9) + 0.01 * omega*1e-9 * 1/np.tanh(hbar * omega / (2 * k * T)))
        
        noise_operator = 'd_hamiltonian_d_er'
        noise_channel = 'Andreev'
        
        T = 0.015
        return self._get_t1_vs_paramvals(
            param_name, param_vals, evals_count, spectrum_data, spectral_density, noise_operator, noise_channel,T, **kwargs
        )
        
        
    def _get_t1_vs_paramvals(
        self, 
        param_name: str, 
        param_vals: np.ndarray, 
        evals_count: int, 
        spectrum_data: SpectrumData, 
        spectral_density: Callable, 
        noise_operator: str, 
        noise_channel: str,
        T: float, 
        **kwargs
        ) -> SpectrumData:
        """
        General method to calculate T1 times for a given noise channel over a range of parameter values.
        
        Parameters
        ----------
        param_name : str
            The name of the parameter to vary.
        param_vals : np.ndarray
            The values of the parameter to vary.
        evals_count : int
            The number of eigenvalues and eigenstates to calculate.
        spectrum_data : SpectrumData
            Precomputed spectral data to use.
        spectral_density : Callable
            Function to calculate the spectral density.
        noise_operator : str
            The noise operator to use ('n_operator' or 'phase_operator').
        noise_channel : str
            The noise channel to use ('capacitive' or 'inductive').
        T : float
            The temperature.
        **kwargs
            Additional arguments to pass to the T1 calculation method.

        Returns
        -------
        SpectrumData
            The T1 times for the specified noise channel over the range of parameter values.
        """
        if spectrum_data is None or noise_operator not in spectrum_data.matrixelem_table:
            spectrum_data = self.get_matelements_vs_paramvals(noise_operator, param_name, param_vals, evals_count=evals_count)
            
        evals_array = spectrum_data.energy_table
        transition_table = evals_array[:, :, np.newaxis] - evals_array[:, np.newaxis, :]
        
        omega = 2 * np.pi * transition_table * 1e9
        epsilon = 1e-12  # Pequeña constante para evitar divisiones por cero
        omega = np.where(omega == 0, epsilon, omega) 
        
        s = spectral_density(omega, T)
        
        matrix_element = spectrum_data.matrixelem_table[noise_operator]
        rate = 2 * np.pi * np.abs(matrix_element)**2 * s
        
        rate *= 1e9  # Convert to rad/s
        rate = np.where(rate == 0, epsilon, rate)
        t1_table = 1 / rate
        
        for idx in range(t1_table.shape[0]):
            np.fill_diagonal(t1_table[idx], np.nan)
        
        spectrum_data.t1_table[noise_channel] = t1_table        
        return spectrum_data
    
    def _get_tphi_vs_paramvals(
        self, 
        param_name: str, 
        param_vals: np.ndarray, 
        A_noise: float,
        noise_channel: str,
        noise_operator: str,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> SpectrumData:
        """
        Calculates the Tphi times for given noise channels over a range of parameter values.

        Parameters
        ----------
        param_name : str
            The name of the parameter to vary.
        param_vals : np.ndarray
            The values of the parameter to vary.
        A_noise : float
            The amplitude of the noise.
        noise_channel : str
            The noise channel to calculate ('flux', etc.).
        noise_operator : str
            The noise operator to use.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments to pass to the Tphi calculation method.

        Returns
        -------
        SpectrumData
            The Tphi times for the specified noise channels over the range of parameter values.
        """
        p = {"omega_ir": 2 * np.pi * 1, "omega_uv": 3 * 2 * np.pi * 1e6, "t_exp": 10e-6}
        p.update(kwargs)
            
        if spectrum_data is None or noise_operator not in spectrum_data.matrixelem_table:
            new_spec = self.get_matelements_vs_paramvals(noise_operator, param_name, param_vals, evals_count=evals_count)
            spectrum_data.matrixelem_table[noise_operator] = new_spec.matrixelem_table[noise_operator]
            
        param_vals = spectrum_data.param_vals
                            
        dE_d_lambda = np.diagonal(spectrum_data.matrixelem_table[noise_operator], axis1=1, axis2 = 2)
        dEij_d_lambda = dE_d_lambda[:,:,np.newaxis] - dE_d_lambda[:,np.newaxis,:]
        
        if spectrum_data.param_name == "phase":
            d2Eij_d_lambda2 = np.gradient(dEij_d_lambda, param_vals, axis=0, edge_order=2)
            rate_1er = np.abs(dEij_d_lambda) * A_noise * np.sqrt(2 * np.abs(np.log(p["omega_ir"] * p["t_exp"])))
            rate_2nd = np.abs(d2Eij_d_lambda2) * A_noise**2  * np.sqrt(2 * np.log(p['omega_uv']/p['omega_ir'])**2 + 2 * np.log(p["omega_ir"] * p["t_exp"])**2)
            
            rate = np.sqrt(rate_1er**2 + rate_2nd**2)
        else:
            rate = np.abs(dEij_d_lambda) * A_noise * np.sqrt(2 * np.abs(np.log(p["omega_ir"] * p["t_exp"])))
        
        epsilon = 1e-12  # Pequeña constante para evitar divisiones por cero
        rate = np.where(rate == 0, epsilon, rate) 
        rate *= 2 * np.pi * 1e9 # Convert to rad/s
        tphi_table = 1 / rate
        
        for idx in range(tphi_table.shape[0]):
            np.fill_diagonal(tphi_table[idx], np.nan)
            
        spectrum_data.tphi_table[noise_channel] = tphi_table
        
        return spectrum_data

    def get_tphi_flux_vs_paramvals(
        self, 
        param_name: str, 
        param_vals: np.ndarray, 
        A_noise: float = 1e-6,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> SpectrumData:
        """
        Calculates the Tphi times for flux noise over a range of parameter values.

        Parameters
        ----------
        param_name : str
            The name of the parameter to vary.
        param_vals : np.ndarray
            The values of the parameter to vary.
        A_noise : float
            The amplitude of the noise.
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments to pass to the Tphi calculation method.

        Returns
        -------
        SpectrumData
            The Tphi times for flux noise over the range of parameter values.
        """
        return self._get_tphi_vs_paramvals(
            param_name=param_name, 
            param_vals=param_vals, 
            A_noise=A_noise, 
            noise_channel='flux_noise', 
            noise_operator='d_hamiltonian_d_phase', 
            evals_count=evals_count, 
            spectrum_data=spectrum_data, 
            **kwargs
        )
        
    def get_tphi_charge_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        A_noise: float = 1e-4,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the Tphi times for charge noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        A_noise : float, optional
            The amplitude of the noise (default is 1e-4).
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments to pass to the Tphi calculation method.
            
        Returns
        -------
        SpectrumData
            The Tphi times for charge noise over the range of parameter values.
        """
        return self._get_tphi_vs_paramvals(
            param_name=param_name, 
            param_vals=param_vals, 
            A_noise=A_noise, 
            noise_channel='charge_noise', 
            noise_operator='d_hamiltonian_d_ng', 
            evals_count=evals_count, 
            spectrum_data=spectrum_data, 
            **kwargs
        )
        
    def get_tphi_CQPS_vs_paramvals(
        self,
        param_name: str = None,
        param_vals: np.ndarray = None,
        fp: float = 17e9,
        z: float = 0.05,
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
    ) -> SpectrumData:
        """
        Calculates the Tphi times for Coherence Quantum Phase Slip noise over a range of parameter values.
        
        Parameters
        ----------
        param_name : str, optional
            The name of the parameter to vary.
        param_vals : np.ndarray, optional
            The values of the parameter to vary.
        fp : float, optional
            The plasma frequency of the junctions in the array (default is 17 GHz).
        z : float, optional
            The adimensional impedance z = Z / R_Q (default is 0.05).
        evals_count : int, optional
            The number of eigenvalues and eigenstates to calculate (default is 6).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments to pass to the Tphi calculation method.
            
        Returns
        -------
        SpectrumData
            The Tphi times for Coherence Quantum Phase Slip noise over the range of parameter values.
        """
        
        if spectrum_data is None:
            spectrum_data = self.get_matelements_vs_paramvals('displacement_operator', param_name, param_vals, evals_count=evals_count)
        elif 'displacement_operator' not in spectrum_data.matrixelem_table:
            new_spec = self.get_matelements_vs_paramvals(
                'displacement_operator',
                spectrum_data.param_name,
                spectrum_data.param_vals,
                evals_count=evals_count
            )
            spectrum_data.matrixelem_table['displacement_operator'] = new_spec.matrixelem_table['displacement_operator']            
        phase_slip_frequency = 4 * np.sqrt(2)/np.pi * fp / np.sqrt(z) * np.exp(-4 / np.pi / z)
        
        displacement_operator = spectrum_data.matrixelem_table['displacement_operator']
        displacement_operator_diagonal = np.diagonal(displacement_operator, axis1=1, axis2=2)
        
        structure_factor = displacement_operator_diagonal[:,:,np.newaxis] - displacement_operator_diagonal[:,np.newaxis,:]
        
        if param_name == 'El':
            N_junctions = fp / 2 / np.pi / (param_vals * 1e9) / z
        else:
            N_junctions  = fp / 2 / np.pi / (self.El * 1e9) / z
        
        rate = np.pi * np.sqrt(N_junctions) * phase_slip_frequency * np.abs(structure_factor)
        rate[rate == 0] = np.inf
        tphi_table = 1 / rate
        
        noise_channel = 'CQPS'
        
        spectrum_data.tphi_table[noise_channel] = tphi_table
        
        return spectrum_data
        
    
    def get_tphi_vs_paramvals(
        self, 
        noise_channels: Union[str, List[str]],
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        evals_count: int = 6,
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> SpectrumData:
        
        if spectrum_data is not None:
            param_name = spectrum_data.param_name
            param_vals = spectrum_data.param_vals
            evals_count = spectrum_data.energy_table.shape[1]
        elif param_name is None or param_vals is None:
            raise ValueError("If spectrum_data is None, param_name and param_vals must be provided.")
        
        if isinstance(noise_channels, str):
            noise_channels = [noise_channels]
                        
        if 'flux_noise' in noise_channels:
            A_noise = kwargs.pop('A_noise', 1e-6)
            spectrum_data = self.get_tphi_flux_vs_paramvals(param_name, param_vals, A_noise, evals_count, spectrum_data, **kwargs)
            
        if 'charge_noise' in noise_channels:
            A_noise = kwargs.pop('A_noise', 1e-4)
            spectrum_data = self.get_tphi_charge_vs_paramvals(param_name, param_vals, A_noise, evals_count, spectrum_data, **kwargs)
            
        if 'CQPS' in noise_channels:
            fp = kwargs.pop('fp', 17e9)
            z = kwargs.pop('z', 0.05)
            spectrum_data = self.get_tphi_CQPS_vs_paramvals(param_name, param_vals, fp, z, evals_count, spectrum_data, **kwargs)
            
        return spectrum_data
    
    def plot_evals_vs_paramvals(
        self, 
        param_name: str = None,
        param_vals: np.ndarray = None,
        evals_count: int = 6,
        subtract_ground: bool = False,
        spectrum_data: SpectrumData = None, 
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot eigenvalues as a function of a parameter.

        Parameters
        ----------
        param_name : str
            Name of the parameter to vary.
        param_vals : np.ndarray
            Values of the parameter to vary.
        evals_count : int, optional
            Number of eigenvalues to calculate (default is 6).
        subtract_ground : bool, optional
            Whether to subtract the ground state energy from all eigenvalues (default is False).
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.
            - color: str or list of str, optional
                Color of the lines. Can be a single color or a list of colors.
            - linestyle: str or list of str, optional
                Linestyle of the lines. Can be a single linestyle or a list of linestyles.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if spectrum_data is None:
            if param_name is None or param_vals is None:
                raise ValueError("Both param_name and param_vals must be provided.")
            spectrum_data = self.get_spectrum_vs_paramvals(param_name, param_vals, evals_count=evals_count)
        else:
            param_name = spectrum_data.param_name
            param_vals = spectrum_data.param_vals
            
        evals = spectrum_data.energy_table.copy()
        if subtract_ground:
            evals -= evals[:, 0][:, np.newaxis]
            evals = evals[:, 1:] # Remove ground state

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle(param_name))
        else:
            fig, ax = fig_ax

        if param_name == 'phase':
            param_vals = param_vals/2/np.pi
            
        color = kwargs.get("color", None)
        linestyle = kwargs.get("linestyle", None)
        ax.plot(param_vals, evals[:,:evals_count], color=color, linestyle=linestyle)

        xlabel = self.PARAM_LABELS.get(param_name, param_name)
        
        ax.set_xlabel(xlabel)
        ax.set_ylabel("Energy [GHz]")
        # ax.grid(True)

        return fig, ax
    
    def plot_matelem_vs_paramvals(
        self, 
        operator: str, 
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        select_elems: Union[int, List[Tuple[int, int]]] = [(1,0)], 
        mode: str = "abs",
        spectrum_data: SpectrumData = None, 
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot matrix elements as a function of a parameter.

        Parameters
        ----------
        operator : str
            Name of the operator.
        param_name : str
            Name of the parameter to vary.
        param_vals : np.ndarray
            Values of the parameter to vary.
        select_elems : Union[int, List[Tuple[int, int]]], optional
            Number of elements to select or list of specific elements to plot (default is 4).
        mode : str, optional
            Mode for plotting the matrix elements ('abs', 'real', 'imag') (default is 'abs').
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if spectrum_data is None:
            if param_name is None or param_vals is None:
                raise ValueError("Both param_name and param_vals must be provided.")
            select_elems = select_elems if isinstance(select_elems, list) else [select_elems]
            evals_count = max(max(i,j) for i, j in select_elems) + 1
            spectrum_data = self.get_matelements_vs_paramvals(operator, param_name, param_vals, evals_count=evals_count)
        else:
            param_name = spectrum_data.param_name
            param_vals = spectrum_data.param_vals

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle(param_name))
        else:
            fig, ax = fig_ax

        matrixelem_table = spectrum_data.matrixelem_table[operator]

        if isinstance(select_elems, int):
            select_elems = [(i, j) for i in range(select_elems) for j in range(i, select_elems)]

        operator_label = self.OPERATOR_LABELS.get(operator, operator)
        
        if param_name == 'phase':
            param_vals = param_vals/2/np.pi
            
        for (i, j) in select_elems:
            if mode == "abs":
                values = np.abs(matrixelem_table[:, i, j])
                ylabel = rf"$|\langle {{i}} | {operator_label} | {{j}} \rangle|$"
            elif mode == "real":
                values = matrixelem_table[:, i, j].real
                ylabel = rf"$\Re(\langle {{i}} | {operator_label} | {{j}} \rangle)$"
            elif mode == "imag":
                values = matrixelem_table[:, i, j].imag
                ylabel = rf"$\Im(\langle {{i}} | {operator_label} | {{j}} \rangle)$"
            else:
                raise ValueError(f"Unsupported mode: {mode}")

            ax.plot(param_vals, values, label=f"({i},{j})")

        xlabel = self.PARAM_LABELS.get(param_name, param_name)
        
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.legend()
        # ax.grid(True)

        return fig, ax
    
    def plot_t1_vs_paramvals(
        self,
        noise_channels: List[str],
        param_name: str = None, 
        param_vals: np.ndarray = None, 
        select_elems: Union[int, List[Tuple[int, int]]] = [(1, 0)],
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot T1 times due to capacitive noise as a function of a parameter.

        Parameters
        ----------
        param_name : str
            Name of the parameter to vary.
        param_vals : np.ndarray
            Values of the parameter to vary.
        select_elems : Union[int, List[Tuple[int, int]]], optional
            Number of elements to select or list of specific elements to plot (default is 4).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if isinstance(noise_channels, str):
            noise_channels = [noise_channels]
            
        if isinstance(select_elems, int):
            evals_count = select_elems
        elif isinstance(select_elems, list):
            evals_count = max(max(i,j) for i, j in select_elems) + 1
                        
        if spectrum_data is None or spectrum_data.t1_table is None or not all(channel in spectrum_data.t1_table for channel in noise_channels):
            spectrum_data = self.get_t1_vs_paramvals(noise_channels, param_name, param_vals, evals_count=evals_count, spectrum_data=spectrum_data)

        param_name = spectrum_data.param_name
        param_vals = spectrum_data.param_vals
            
        if isinstance(select_elems, int):
            select_elems = [(i, j) for i in range(select_elems) for j in range(i) if i > j]

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle(param_name))
        else:
            fig, ax = fig_ax
            
        rate_effective = np.zeros_like(param_vals)
        
        if param_name == 'phase':
            param_vals = param_vals/2/np.pi

        for (i, j) in select_elems:
            for channel in noise_channels:
                t1_times = spectrum_data.t1_table[channel][:, i, j]
                rate_effective += 1 / t1_times
                ax.plot(param_vals, t1_times, label=f"{channel} ({i},{j})")

        ax.plot(param_vals, 1/rate_effective, label=f"T1 effective", color='black', linestyle='--')

        xlabel = self.PARAM_LABELS.get(param_name, param_name)
        
        ax.set_xlabel(xlabel)
        ax.set_ylabel(r"$T_1$ [s]")
        ax.legend()
        # ax.grid(True)

        return fig, ax
    
    def plot_tphi_vs_paramvals(
        self, 
        noise_channels: List[str], 
        param_name: str = None, 
        param_vals: np.ndarray = None,
        select_elems: Union[int, List[Tuple[int, int]]] = [(1,0)], 
        spectrum_data: SpectrumData = None,
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot Tphi times due to specified noise channels as a function of a parameter.

        Parameters
        ----------
        param_name : str
            Name of the parameter to vary.
        param_vals : np.ndarray
            Values of the parameter to vary.
        select_elems : Union[int, List[Tuple[int, int]]] optional
            Number of elements to select or list of specific elements to plot (default is 4).
        noise_channels : List[str], optional
            List of noise channels to consider (default is ['capacitive']).
        spectrum_data : SpectrumData, optional
            Precomputed spectral data to use (default is None).
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if spectrum_data is None:
            evals_count = max(max(i,j) for i, j in select_elems) + 1
            operators = set()
            for channel in noise_channels:
                if channel == 'flux_noise':
                    operators.add('d_hamiltonian_d_phase')
                if channel == 'charge_noise':
                    operators.add('d_hamiltonian_d_ng')
                else:
                    raise ValueError(f"Unsupported Tphi noise channel: {channel}")
            spectrum_data = self.get_matelements_vs_paramvals(list(operators), param_name, param_vals, evals_count=evals_count)
        else:
            param_name = spectrum_data.param_name
            param_vals = spectrum_data.param_vals

        if isinstance(select_elems, int):
            select_elems = [(i, j) for i in range(select_elems) for j in range(i) if i > j]

        if isinstance(noise_channels, str):
            noise_channels = [noise_channels]

        for channel in noise_channels:
            if channel not in spectrum_data.tphi_table:
                spectrum_data = self.get_tphi_vs_paramvals([channel], param_name, param_vals, spectrum_data=spectrum_data)
                
        if param_name == 'phase':
            param_vals = param_vals/2/np.pi

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle(param_name))
        else:
            fig, ax = fig_ax
            
        for (i, j) in select_elems:
            for channel in noise_channels:
                tphi_times = spectrum_data.tphi_table[channel][:,i,j]
                ax.plot(param_vals, tphi_times, label=f"Tphi {channel} ({i}->{j})")            

        xlabel = self.PARAM_LABELS.get(param_name, param_name)
        
        ax.set_xlabel(xlabel)
        ax.set_ylabel(r"$T_\varphi$ [s]")
        ax.legend()

        return fig, ax  
    
    def set_param(self, param_name: str, val: Any) -> None:
        """
        Sets the value of a parameter if it exists.

        Parameters
        ----------
        param_name : str
            The name of the parameter to set.
        val : Any
            The value to set for the parameter.

        Raises
        ------
        AttributeError
            If the parameter does not exist.
        """
        if not hasattr(self, param_name):
            raise AttributeError(f"Parameter '{param_name}' does not exist in the Ferbo class.")
        setattr(self, param_name, val)
        
    def _generate_suptitle(self, exclude_params: Union[str, List[str]] = None) -> str:
        """
        Generate the suptitle for the plot, excluding the specified parameters.

        Parameters
        ----------
        exclude_params : Union[str, List[str]]
            The parameter(s) to exclude from the title.

        Returns
        -------
        str
            The generated suptitle.
        """
        if isinstance(exclude_params, str):
            exclude_params = [exclude_params]
        exclude_params = exclude_params or []

        def format_value(value):
            if isinstance(value, float):
                return f'{value:.4f}'.rstrip('0').rstrip('.')
            return str(value)

        title_parts = [
            rf'{label} = {format_value(getattr(self, param))}' if param not in exclude_params else ''
            for param, label in self.PARAM_LABELS.items()
        ]
        return ', '.join(filter(None, title_parts))


import numpy as np
from .qubit_base import QubitBase
from typing import Union, Tuple, Dict, Any, Iterable, Optional
import matplotlib.pyplot as plt
from scipy.integrate import quad
from .utilities import cos_kphi_operator


class Gatemon(QubitBase):
    
    PARAM_LABELS = {
        'Ec': r'$E_C$',
        'Delta': r'$\Delta$',
        'T': r'$T$',
        'ng': r'$n_g$'
        }
    
    OPERATOR_LABELS = {
    'n_operator': r'\hat{n}',
    'phase_operator': r'\hat{\phi}',
    'd_hamiltonian_d_ng': r'\partial \hat{H} / \partial n_g',
    }
    
    def __init__(self, Ec, Delta, T, ng, n_cut):
        
        self.Ec = Ec
        self.Delta = Delta
        self.T = T
        self.ng = ng
        self.n_cut = n_cut
        self.num_coef = 4
        self.dimension = 2 * n_cut + 1
        
        super().__init__(dimension = self.dimension)
        
    def n_operator(self):
        """
        Generate the number operator matrix \hat{n} in a (2N+1)-dimensional truncated space.
        
        Returns:
            numpy.ndarray: Diagonal matrix representing the number operator.
        """
        n_values = np.arange(-self.n_cut, self.n_cut+1)
        return np.diag(n_values)
    
    def junction_potential(self):
        
        def f(phi, T, Delta):
            return -Delta * np.sqrt(1 - T * np.sin(phi/2)**2)

        # Cálculo numérico de A_k para k >= 1
        def A_k(k, T, Delta):
            integral, error = quad(lambda x: f(x, T, Delta) * np.cos(k*x), 0, np.pi)
            return 2 * integral / np.pi
        
        A_coeffs = [A_k(k, self.T, self.Delta) for k in range(self.num_coef + 1)]
        
        junction_term = A_coeffs[0] / 2 * np.eye(self.dimension, dtype=np.complex128)
        for k in range(1, self.num_coef + 1 ):
            junction_term += A_coeffs[k] * cos_kphi_operator(k, self.dimension)
            
        return junction_term
    
    def hamiltonian(self):
        n_op = self.n_operator() - self.ng * np.eye(self.dimension)      
        kinetic_term = 4 * self.Ec * (n_op @ n_op)
        junction_term = self.junction_potential()

        return kinetic_term + junction_term
            
    def potential(self, phi: Union[float, np.ndarray]):
        raise NotImplementedError("Potential method not implemented for this class.")

    def d_hamiltonian_d_ng(self):
        n_op = self.n_operator() - self.ng * np.eye(self.dimension)
        return - 8 * self.Ec * n_op
        
    def wavefunction(self, which: int = 0, phi_grid: np.ndarray = None, esys: Tuple[np.ndarray, np.ndarray] = None) -> Dict[str, Any]:
        raise NotImplementedError("Wavefunction method not implemented for this class.")
    
    def plot_wavefunction(
        self, 
        which: Union[int, Iterable[int]] = 0, 
        phi_grid: np.ndarray = None, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        scaling: Optional[float] = None,
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        raise NotImplementedError("Wavefunction plotting not implemented for this class.")
        

        


import matplotlib.pyplot as plt
import numpy as np
from .qubit_base import QubitBase
from scipy.linalg import cosm, sinm, eigh, expm
from typing import Any, Dict, Optional, Tuple, Union, Iterable
from .operators import destroy, creation, sigma_z, sigma_y, sigma_x

class Ferbo(QubitBase):
    PARAM_LABELS = {
        'Ec': r'$E_C$',
        'El': r'$E_L$',
        'Gamma': r'$\Gamma$',
        'delta_Gamma': r'$\delta \Gamma$',
        'er': r'$\epsilon_r$',
        'phase': r'$\Phi_{\mathrm{ext}} / \Phi_0$'
    }
    
    OPERATOR_LABELS = {
    'n_operator': r'\hat{n}',
    'phase_operator': r'\hat{\phi}',
    'd_hamiltonian_d_ng': r'\partial \hat{H} / \partial n_g',
    'd_hamiltonian_d_phase': r'\partial \hat{H} / \partial \phi_{{ext}}',
    'd_hamiltonian_d_EL': r'\partial \hat{H} / \partial E_L',
    'd_hamiltonian_d_deltaGamma': r'\partial \hat{H} / \partial \delta \Gamma',
    'd_hamiltonian_d_er': r'\partial \hat{H} / \partial \epsilon_r',
    }
    
    def __init__(self, Ec, El, Gamma, delta_Gamma, er, phase, dimension, flux_grouping: str = 'ABS', Delta = 40):
        """
        Initializes the Ferbo class with the given parameters.

        Parameters
        ----------
        Ec : float
            Charging energy.
        El : float
            Inductive energy.
        Gamma : float
            Coupling strength.
        delta_Gamma : float
            Coupling strength difference.
        er : float
            Energy relaxation rate.
        phase : float
            External magnetic phase.
        dimension : int
            Dimension of the Hilbert space.
        flux_grouping : str, optional
            Flux grouping ('L' or 'ABS') (default is 'L').
        Delta : float
            Superconducting gap.
        """
        if flux_grouping not in ['L', 'ABS']:
            raise ValueError("Invalid flux grouping; must be 'L' or 'ABS'.")
        
        self.Ec = Ec
        self.El = El
        self.Gamma = Gamma
        self.delta_Gamma = delta_Gamma
        self.er = er
        self.phase = phase
        self.dimension = dimension // 2 * 2
        self.flux_grouping = flux_grouping
        self.Delta = Delta
        super().__init__(self.dimension)
        
    @property
    def phase_zpf(self) -> float:
        """
        Returns the zero-point fluctuation of the phase.

        Returns
        -------
        float
            Zero-point fluctuation of the phase.
        """
        return (2 * self.Ec / self.El) ** 0.25
    
    @property
    def n_zpf(self) -> float:
        """
        Returns the zero-point fluctuation of the charge number.

        Returns
        -------
        float
            Zero-point fluctuation of the charge number.
        """
        return 1/2 * (self.El / 2 / self.Ec) ** 0.25
    
    def phi_osc(self) -> float:
        """
        Returns the oscillator length for the LC oscillator composed of the inductance and capacitance.

        Returns
        -------
        float
            Oscillator length.
        """
        return (8.0 * self.Ec / self.El) ** 0.25
    
    def n_operator(self) -> np.ndarray:
        """
        Returns the charge number operator.

        Returns
        -------
        np.ndarray
            The charge number operator.
        """
        single_mode_n_operator = 1j * self.n_zpf * (creation(self.dimension //2 ) - destroy(self.dimension // 2))
        return np.kron(np.eye(2), single_mode_n_operator)
    
    def phase_operator(self) -> np.ndarray:
        """
        Returns the total phase operator.

        Returns
        -------
        np.ndarray
            The total phase operator.
        """
        single_mode_phase_operator = self.phase_zpf * (creation(self.dimension //2) + destroy(self.dimension //2))
        return np.kron(np.eye(2), single_mode_phase_operator)        
    
    def jrl_potential(self) -> np.ndarray:
        """
        Returns the Josephson Resonance Level potential.

        Returns
        -------
        np.ndarray
            The Josephson Resonance Level potential.
        """
        phase_op = self.phase_operator()[:self.dimension//2, :self.dimension//2]
        if self.flux_grouping == 'ABS':
            phase_op -= self.phase * np.eye(self.dimension // 2)
        
        return - self.Gamma * np.kron(sigma_z(), cosm(phase_op/2)) \
            - self.delta_Gamma * np.kron(sigma_y(), sinm(phase_op/2)) \
                + self.er * np.kron(sigma_x(), np.eye(self.dimension // 2))

    
    # def zazunov_potential(self) -> np.ndarray:
        
    def hamiltonian(self) -> np.ndarray:
        """
        Returns the Hamiltonian of the system.

        Returns
        -------
        np.ndarray
            The Hamiltonian of the system.
        """
        n_x = self.delta_Gamma/4/(self.Gamma+self.Delta)
        n_op = self.n_operator() + n_x * np.kron(sigma_x(),np.eye(self.dimension//2))
        
        charge_term = 4 * self.Ec * n_op @ n_op
        phase_op = self.phase_operator()
        
        if self.flux_grouping == 'ABS':
            inductive_term = 0.5 * self.El * phase_op @ phase_op
        else:
            phase_op += self.phase * np.eye(self.dimension)
            inductive_term = 0.5 * self.El * phase_op @ phase_op
            
        potential = self.jrl_potential()
        return charge_term + inductive_term + potential
    
    def d_hamiltonian_d_EL(self) -> np.ndarray:
        
        if self.flux_grouping == 'L':
            phase_op = self.phase_operator()
        elif self.flux_grouping == 'ABS':
            phase_op = self.phase_operator() - self.phase * np.eye(self.dimension)

        return 1/2 * np.dot(phase_op, phase_op)
    
    def d_hamiltonian_d_ng(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the number of charge offset.
        
        Returns
        -------
        np.ndarray
            The derivative of the Hamiltonian with respect to the number of charge offset.
        
        """
        return 8 * self.Ec * (self.n_operator() - self.delta_Gamma/4/(self.Gamma+self.Delta) * np.kron(sigma_z(), np.eye(self.dimension//2)))
    
    def d_hamiltonian_d_phase(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the external magnetic phase.

        Returns
        -------
        np.ndarray
            The derivative of the Hamiltonian with respect to the external magnetic phase.
        """
        if self.flux_grouping == 'L':
            return self.El * (self.phase_operator() + self.phase * np.eye(self.dimension))
        elif self.flux_grouping == 'ABS':
            phase_op = self.phase_operator()[:self.dimension//2,:self.dimension//2] - self.phase * np.eye(self.dimension // 2)
            return - self.Gamma/2 * np.kron(sigma_z(), sinm(phase_op/2)) + self.delta_Gamma/2 * np.kron(sigma_y(), cosm(phase_op/2))
        
                
    def d_hamiltonian_d_er(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the energy relaxation rate.

        Returns
        -------
        Qobj
            The derivative of the Hamiltonian with respect to the energy relaxation rate.
        """
        return + np.kron(sigma_x(), np.eye(self.dimension // 2))
    
    def d_hamiltonian_d_deltaGamma(self) -> np.ndarray:
        """
        Returns the derivative of the Hamiltonian with respect to the coupling strength difference.

        Returns
        -------
        Qobj
            The derivative of the Hamiltonian with respect to the coupling strength difference.
        """
        if self.flux_grouping == 'L':
            phase_op = self.phase_operator()[:self.dimension//2,:self.dimension//2]
        else:
            raise NotImplementedError("Not implemented for ABS grouping.")
            phase_op = self.phase_operator()[:self.dimension//2,:self.dimension//2] - self.phase * np.eye(self.dimension // 2)
        return - np.kron(sigma_y(), sinm(phase_op/2))
    
    def wavefunction(
        self, 
        which: int = 0,
        phi_grid: np.ndarray = None,
        esys: Tuple[np.ndarray, np.ndarray] = None,
        basis: str = 'phase',
        rotate: str = False,
        ) -> Dict[str, Any]:
        """
        Returns a wave function in the phi basis.

        Parameters
        ----------
        which : int, optional
            Index of desired wave function (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.
        basis : str, optional
            Basis in which to return the wavefunction ('phase' or 'charge') (default is 'phase').
        rotate : bool, optional
            Whether to rotate the basis (default is False).
        Returns
        -------
        Dict[str, Any]
            Wave function data containing basis labels, amplitudes, and energy.
        """
        if esys is None:
            evals_count = max(which + 1, 3)
            evals, evecs = self.eigensys(evals_count)
        else:
            evals, evecs = esys
            
        dim = self.dimension//2
                
        if rotate:
            
            I = np.eye(self.dimension//2)
            change_of_basis_operator = (1/np.sqrt(2)) * np.block([[I, I],
                                        [I, -I]])
    
            evecs = (change_of_basis_operator @ evecs)
        
        evecs = evecs.T
        
        if basis == 'phase':
            l_osc = self.phase_zpf
        elif basis == 'charge':
            l_osc = self.n_zpf     
                                
        if phi_grid is None:
            phi_grid = np.linspace(-5 * np.pi, 5 * np.pi, 151)

        phi_basis_labels = phi_grid
        wavefunc_osc_basis_amplitudes = evecs[which, :]
        phi_wavefunc_amplitudes = np.zeros((2, len(phi_grid)), dtype=np.complex128)
        
        for n in range(dim):
            phi_wavefunc_amplitudes[0] += wavefunc_osc_basis_amplitudes[n] * self.harm_osc_wavefunction(n, phi_basis_labels, l_osc)
            phi_wavefunc_amplitudes[1] += wavefunc_osc_basis_amplitudes[self.dimension//2 + n] * self.harm_osc_wavefunction(n, phi_basis_labels, l_osc)

        return {
            "basis_labels": phi_basis_labels,
            "amplitudes": phi_wavefunc_amplitudes,
            "energy": evals[which]
        }
        
                
    def potential(self, phi: Union[float, np.ndarray]) -> np.ndarray:
        """
        Calculates the potential energy for given values of phi.

        Parameters
        ----------
        phi : Union[float, np.ndarray]
            The phase values at which to calculate the potential.

        Returns
        -------
        np.ndarray
            The potential energy values.
        """
        phi_array = np.atleast_1d(phi)
        evals_array = np.zeros((len(phi_array), 2))
        phi_ext = 2 * np.pi * self.phase

        for i, phi_val in enumerate(phi_array):
            if self.flux_grouping == 'ABS':
                inductive_term = 0.5 * self.El * phi_val**2 * np.eye(2)
                andreev_term = -self.Gamma * np.cos((phi_val + self.phase) / 2) * sigma_z() - self.delta_Gamma * np.sin((phi_val + self.phase) / 2) * sigma_y() + self.er * sigma_x()
            elif self.flux_grouping == 'L':
                inductive_term = 0.5 * self.El * (phi_val + phi_ext)**2 * np.eye(2)
                andreev_term = -self.Gamma * np.cos(phi_val / 2) * sigma_z() - self.delta_Gamma * np.sin(phi_val / 2) * sigma_y() + self.er * sigma_x()
            
            potential_operator = inductive_term + andreev_term
            evals_array[i] = eigh(
                potential_operator,
                eigvals_only=True,
                check_finite=False,
        )

        return evals_array
    
    def tphi_1_over_f(
        self, 
        A_noise: float, 
        i: int, 
        j: int, 
        noise_op: str,
        esys: Tuple[np.ndarray, np.ndarray] = None,
        get_rate: bool = False,
        **kwargs
        ) -> float:
        """
        Calculates the 1/f dephasing time (or rate) due to an arbitrary noise source.

        Parameters
        ----------
        A_noise : float
            Noise strength.
        i : int
            State index that along with j defines a qubit.
        j : int
            State index that along with i defines a qubit.
        noise_op : str
            Name of the noise operator, typically Hamiltonian derivative w.r.t. noisy parameter.
        esys : Tuple[np.ndarray, np.ndarray], optional
            Precomputed eigenvalues and eigenvectors (default is None).
        get_rate : bool, optional
            Whether to return the rate instead of the Tphi time (default is False).

        Returns
        -------
        float
            The 1/f dephasing time (or rate).
        """
        p = {"omega_ir": 2 * np.pi * 1, "omega_uv": 3 * 2 * np.pi * 1e6, "t_exp": 10e-6}
        p.update(kwargs)
                
        if esys is None:
            evals, evecs = self.eigensys(evals_count=max(j, i) + 1)
        else:
            evals, evecs = esys

        noise_operator = getattr(self, noise_op)()    
        dEij_d_lambda = np.abs(evecs[i].conj().T @ noise_operator @ evecs[i] - evecs[j].conj().T @ noise_operator @ evecs[j])

        rate = (dEij_d_lambda * A_noise * np.sqrt(2 * np.abs(np.log(p["omega_ir"] * p["t_exp"]))))
        rate *= 2 * np.pi * 1e9 # Convert to rad/s

        return rate if get_rate else 1 / rate
    
    def tphi_1_over_f_flux(
        self, 
        A_noise: float = 1e-6,
        i: int = 0, 
        j: int = 1, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        get_rate: bool = False, 
        **kwargs
        ) -> float:
        return self.tphi_1_over_f(A_noise, i, j, 'd_hamiltonian_d_phase', esys=esys, get_rate=get_rate, **kwargs)

    def plot_wavefunction(
        self, 
        which: Union[int, Iterable[int]] = 0, 
        phi_grid: np.ndarray = None, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        scaling: Optional[float] = 1,
        plot_potential: bool = False,
        basis: str = 'phase',
        rotate: bool = False,
        mode: str = 'abs',
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the wave function in the phi basis.

        Parameters
        ----------
        which : Union[int, Iterable[int]], optional
            Index or indices of desired wave function(s) (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.
        esys : Tuple[np.ndarray, np.ndarray], optional
            Precomputed eigenvalues and eigenvectors.
        scaling : float, optional
            Scaling factor for the wavefunction (default is 1).
        plot_potential : bool, optional
            Whether to plot the potential (default is False).
        basis: str, optional
            Basis in which to return the wavefunction ('phase' or 'charge') (default is 'phase').
        rotate : bool, optional
            Whether to rotate the basis (default is False).
        mode: str, optional
            Mode of the wavefunction ('abs', 'real', or 'imag') (default is 'abs').
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if isinstance(which, int):
            which = [which]
            
        if phi_grid is None:
            phi_grid = np.linspace(-5 * np.pi, 5 * np.pi, 151)
            
        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle())
        else:
            fig, ax = fig_ax
        
        if plot_potential:
            potential = self.potential(phi=phi_grid)
            ax.plot(phi_grid, potential[:, 0], color='black', label='Potential')
            ax.plot(phi_grid, potential[:, 1], color='black')

        for idx in which:
            wavefunc_data = self.wavefunction(which=idx, phi_grid=phi_grid, esys=esys, basis=basis, rotate=rotate)
            phi_basis_labels = wavefunc_data["basis_labels"]
            wavefunc_amplitudes = wavefunc_data["amplitudes"]
            wavefunc_energy = wavefunc_data["energy"]
            
            if mode == 'abs':
                y_values = np.abs(wavefunc_amplitudes[0])
                y_values_down = np.abs(wavefunc_amplitudes[1])
            elif mode == 'real':
                y_values = wavefunc_amplitudes[0].real
                y_values_down = wavefunc_amplitudes[1].real
            elif mode == 'imag':
                y_values = wavefunc_amplitudes[0].imag
                y_values_down = wavefunc_amplitudes[1].imag
            else:
                raise ValueError("Invalid mode; must be 'abs', 'real', or 'imag'.")

            ax.plot(
                phi_basis_labels,
                wavefunc_energy + scaling * y_values,
                label=rf"$\Psi_{idx} \uparrow $"
                )
            ax.plot(
                phi_basis_labels, 
                wavefunc_energy + scaling * y_values_down,
                label=rf"$\Psi_{idx} \downarrow $"
                )
            
        if basis == 'phase':
            ax.set_xlabel(r"$2 \pi \Phi / \Phi_0$")
            ax.set_ylabel(r"$\psi(\varphi)$, Energy [GHz]")
        elif basis == 'charge':
            ax.set_xlabel(r"$n$")
            ax.set_ylabel(r"$\psi(n)$, Energy [GHz]")
        
        ax.legend()
        ax.grid(True)

        return fig, ax      


import numpy as np

def sin_kphi_operator(k: int, dimension: int, phi_ext: float = 0) -> np.ndarray:
    """
    Generate the matrix representation of the \\sin(k\\hat{\\phi}) operator in the number basis.

    The operator is defined via the exponential representation:
        sin(k\\phi) = (e^(ik\\phi) - e^(-ik\\phi))/(2i)
    In the number basis, the matrix element corresponding to a shift by +k is 1/(2i)
    and by -k is -1/(2i). Note that 1/(2i) equals -0.5j.

    Parameters
    ----------
    k : int
        The integer multiplier of \\hat{\\phi}.
    dimension : int
        Dimension of the Hilbert space.
    phi_ext : float, optional
        External phase offset, by default 0.

    Returns
    -------
    numpy.ndarray
        Matrix representation of \\sin(k\\hat{\\phi}).

    Notes
    -----
    When k == 0, sin(0) = 0, so the operator is the zero operator.
    """
    if k == 0:
        return np.zeros((dimension, dimension))
    
    sin_kphi = np.zeros((dimension, dimension), dtype=complex)
    indices = np.arange(dimension)

    # For terms <n|exp(i k phi)|m>: m = n + k
    mask_up = indices + k < dimension
    # For terms <n|exp(-i k phi)|m>: m = n - k
    mask_down = indices - k >= 0

    # According to the definition:
    # sin(k phi) = [exp(i k phi) - exp(-i k phi)]/(2i)
    # 1/(2i) = -0.5j, so:
    sin_kphi[indices[mask_up], indices[mask_up] + k] = -0.5j * np.exp(-1j * phi_ext)
    sin_kphi[indices[mask_down], indices[mask_down] - k] = 0.5j * np.exp(1j * phi_ext)

    return sin_kphi

def cos_kphi_operator(k:int, dimension: int, phase: float = 0) -> np.ndarray:
    """
    Generate the matrix representation of the \cos(k\hat{\phi}) operator in the number basis.
    
    Parameters:
        k (int): The integer multiplier of \hat{\phi}.
        dimension (int): Dimension of the Hilbert space.
        phase (float, optional): Phase offset, by default
    
    Returns:
        numpy.ndarray: Matrix representation of \cos(k\hat{\phi}).
    """
    
    if k == 0:
        return np.eye(dimension)
    
    cos_kphi = np.zeros((dimension, dimension), dtype=np.complex128)
    indices = np.arange(dimension)
    
    mask_up = indices + k < dimension
    mask_down = indices - k >= 0
    
    cos_kphi[indices[mask_up], indices[mask_up] + k] = 0.5 * np.exp( - 1j * phase)
    cos_kphi[indices[mask_down], indices[mask_down] - k] = 0.5 * np.exp(1j * phase)
    
    return cos_kphi


from qutip import Qobj
import scipy.sparse as sp
import numpy as np

def state_to_density_matrix(state_vector: np.ndarray) -> np.ndarray:
    """
    Convert a state vector to a density matrix.

    Parameters
    ----------
    state_vector : numpy.ndarray
        The state vector to be converted.

    Returns
    -------
    numpy.ndarray
        The density matrix.
    """
    return np.outer(state_vector, state_vector.conj())

def ptrace(rho: np.ndarray, dims: tuple, subsys: int) -> np.ndarray:
    """
    Compute the partial trace of a density matrix over a specified subsystem.

    The partial trace is a method used in quantum mechanics to obtain the reduced 
    density matrix of a subsystem by tracing out the degrees of freedom of the 
    other subsystem.

    Parameters
    ----------
    rho : numpy.ndarray
        The density matrix to be traced. It should be a square matrix of shape 
        (dimA * dimB, dimA * dimB).
    dims : tuple
        A tuple (dimA, dimB) specifying the dimensions of the subsystems A and B.
    subsys : int
        The subsystem to trace out. Use 0 to trace out subsystem A and 1 to trace 
        out subsystem B.

    Returns
    -------
    numpy.ndarray
        The reduced density matrix after tracing out the specified subsystem.

    Raises
    ------
    ValueError
        If the subsys parameter is not 0 or 1.
    """
    dimA, dimB = dims
    rho_reshaped = rho.reshape([dimA, dimB, dimA, dimB])
    if subsys == 0:
        return np.trace(rho_reshaped, axis1=0, axis2=2)
    elif subsys == 1:
        return np.trace(rho_reshaped, axis1=1, axis2=3)
    else:
        raise ValueError("subsys must be 0 (to trace out A) or 1 (to trace out B).")

def purity(density_matrix: np.ndarray) -> float:
    """
    Calculate the purity of a quantum state.

    The purity is defined as Tr(rho^2), where rho is the density matrix.
    For a pure state, the purity is 1. For a mixed state of dimension `d`, 1/d <= purity < 1.

    Parameters
    ----------
    density_matrix : numpy.ndarray
        The density matrix of the quantum state.

    Returns
    -------
    float
        The purity of the quantum state.
    """
    return np.trace(np.dot(density_matrix, density_matrix)).real

def destroy(dimension: int) -> np.ndarray:
    """
    Returns the annihilation (lowering) operator for a given dimension.

    Parameters
    ----------
    dimension : int
        Dimension of the Hilbert space.

    Returns
    -------
    np.ndarray
        The annihilation operator.
    """
    indices = np.arange(1, dimension)
    data = np.sqrt(indices)
    return np.diag(data, k=1)

def creation(dimension: int) -> np.ndarray:
    """
    Returns the creation (raising) operator for a given dimension.

    Parameters
    ----------
    dimension : int
        Dimension of the Hilbert space.

    Returns
    -------
    np.ndarray
        The creation operator.
    """
    return destroy(dimension).T.conj()

def sigma_x() -> np.ndarray:
    """
    Returns the Pauli-X (sigma_x) operator.

    Returns
    -------
    np.ndarray
        The Pauli-X operator.
    """
    return np.array([[0, 1], [1, 0]], dtype=complex)

def sigma_y() -> np.ndarray:
    """
    Returns the Pauli-Y (sigma_y) operator.

    Returns
    -------
    np.ndarray
        The Pauli-Y operator.
    """
    return np.array([[0, -1j], [1j, 0]], dtype=complex)

def sigma_z() -> np.ndarray:
    """
    Returns the Pauli-Z (sigma_z) operator.

    Returns
    -------
    np.ndarray
        The Pauli-Z operator.
    """
    return np.array([[1, 0], [0, -1]], dtype=complex)

###########

def cos_phi(N, phi_ext, m = 1):
    """
    Compute the cosine phi operator matrix in a complex sparse representation.

    The operator is calculated based on the given system size N and external phase factor phi_ext. 

    Parameters
    ----------
    N : int
        The size of the matrix to be created.
    phi_ext : float
        The external phase factor.
    m : int, optional
        The diagonal offset. Default is 1.

    Returns
    -------
    Qobj
        The cosine phi operator represented as a QuTiP Qobj with the CSR sparse matrix format.
    """
    diags = [np.exp(1j*phi_ext/2)*np.ones(N-m,dtype=int),np.exp(-1j*phi_ext/2)*np.ones(N-m,dtype=int)]
    T = sp.diags(diags,[m,-m],format='csr', dtype=complex)
    return Qobj(T, isherm=True)/2

def sin_phi(N, phi_ext, m = 1):
    """
    Compute the sine phi operator matrix in a complex sparse representation.

    The operator is calculated based on the given system size N and external phase factor phi_ext.

    Parameters
    ----------
    N : int
        The size of the matrix to be created.
    phi_ext : float
        The external phase factor.
    m : int, optional
        The diagonal offset. Default is 1.

    Returns
    -------
    Qobj
        The sine phi operator represented as a QuTiP Qobj with the CSR sparse matrix format.
    """
    diags = [np.exp(1j*phi_ext/2)*np.ones(N-m,dtype=int),-np.exp(-1j*phi_ext/2)*np.ones(N-m,dtype=int)]
    T = sp.diags(diags,[m,-m],format='csr', dtype=complex)
    return Qobj(T, isherm=True) /2/1j 


import numpy as np
from .qubit_base import QubitBase
from scipy.linalg import sinm, cosm
from .operators import destroy, creation
from typing import Union, Tuple, Dict, Any, Iterable, Optional
import matplotlib.pyplot as plt
from scipy.integrate import quad



class Gatemonium(QubitBase):
    
    PARAM_LABELS = {
        'Ec': r'$E_C$',
        'El': r'$E_L$',
        'Delta': r'$\Delta$',
        'T': r'$T$',
        'phase': r'$\Phi_{ext} / \Phi_0$'
    }
    
    OPERATOR_LABELS = {
    'n_operator': r'\hat{n}',
    'phase_operator': r'\hat{\phi}',
    'd_hamiltonian_d_ng': r'\partial \hat{H} / \partial n_g',
    'd_hamiltonian_d_phase': r'\partial \hat{H} / \partial \phi_{{ext}}',
    }
    
    def __init__(self, Ec, El, Delta, T, phase, dimension, flux_grouping):
        
        if flux_grouping not in ['L', 'ABS']:
            raise ValueError("Invalid flux grouping; must be 'L' or 'ABS'.")
        
        self.Ec = Ec
        self.El = El
        self.Delta = Delta
        self.T = T
        self.phase = phase
        self.dimension = dimension
        self.flux_grouping = flux_grouping
        self.num_coef = 4
        
        # self.phi_grid = np.linspace(- 8 * np.pi, 8 * np.pi, self.dimension, endpoint=False)
        # self.dphi = self.phi_grid[1] - self.phi_grid[0] 
        
        super().__init__(dimension = dimension)
        
    @property
    def phase_zpf(self) -> float:
        """
        Returns the zero-point fluctuation of the phase.

        Returns
        -------
        float
            Zero-point fluctuation of the phase.
        """
        return (2 * self.Ec / self.El) ** 0.25
    
    @property
    def n_zpf(self) -> float:
        """
        Returns the zero-point fluctuation of the charge number.

        Returns
        -------
        float
            Zero-point fluctuation of the charge number.
        """
        return 1/2 * (self.El / 2 / self.Ec) ** 0.25
        
    def phi_osc(self) -> float:
        """
        Returns the oscillator length for the LC oscillator composed of the inductance and capacitance.

        Returns
        -------
        float
            Oscillator length.
        """
        return (8.0 * self.Ec / self.El) ** 0.25
        
    def n_operator(self):
        dimension = self.dimension
        return 1j/2 * (self.El/2/self.Ec)**0.25 * (creation(dimension) - destroy(dimension))
    
    def phase_operator(self):
        dimension = self.dimension
        return (2*self.Ec/self.El)**0.25 * (creation(dimension) + destroy(dimension))
    
    def junction_potential(self):
        phase_op = self.phase_operator()
        
        if self.flux_grouping == 'ABS':
            phase_op -= self.phase * np.eye(self.dimension)
        
        junction_term = 0
        def f(phi, T, Delta):
            return - Delta * np.sqrt(1 - T * np.sin(phi/2)**2)
        
        def A_k(k, T, Delta):
            integral, error = quad(lambda x: f(x, T, Delta) * np.cos(k*x), 0, np.pi)
            return 2 * integral / np.pi
        
        A_coeffs = [A_k(k, self.T, self.Delta) for k in range(self.num_coef + 1)]
        
        junction_term = A_coeffs[0] / 2  * np.eye(self.dimension)
        for k in range(1, self.num_coef + 1):
            junction_term += A_coeffs[k] * cosm(k * phase_op)            
        return junction_term
    
    def hamiltonian(self):
        n_op = self.n_operator()
        phase_op = self.phase_operator()

        if self.flux_grouping == 'L':
            phase_op += self.phase * np.eye(self.dimension)
        
        kinetic_term = 4 * self.Ec * (n_op @ n_op)
        inductive_term = 0.5 * self.El * (phase_op @ phase_op)
        junction_term = self.junction_potential()

        return kinetic_term + inductive_term + junction_term
    
    def d_hamiltonian_d_phase(self) -> np.ndarray:
        phase_op = self.phase_operator()
        ext_phase = self.phase * np.eye(self.dimension)
        
        if self.flux_grouping == 'L':
            return self.El * (phase_op + ext_phase)
        elif self.flux_grouping == 'ABS':
            phase_op = self.phase_operator() - self.phase * np.eye(self.dimension)
            # numerator = -self.T * self.Delta * sinm(phase_op - ext_phase)
            # sin_op = sinm((phase_op - ext_phase) / 2)
            # denominator = 4 * sqrtm(np.eye(self.dimension) - self.T * sin_op.conj().T @ sin_op)
            
            # return solve(numerator , denominator)
            
            def f(phi, T, Delta):
                return -Delta * np.sqrt(1 - T * np.sin(phi/2)**2)
            
            def A_k(k, T, Delta):
                integral, error = quad(lambda x: f(x, T, Delta) * np.cos(k*x), 0, np.pi)
                return 2 * integral / np.pi
            
            A_coeffs = [A_k(k, self.T, self.Delta) for k in range(0, self.num_coef + 1)]
                    
            dH_dPhi = 0
            for k in range(self.num_coef):
                dH_dPhi += A_coeffs[k] * k * sinm(k * phase_op)
                
            return dH_dPhi
            
            
        
    def potential(self, phi: Union[float, np.ndarray]):
        
        phi_array = np.atleast_1d(phi)
        
        if self.flux_grouping == 'L':
            inductive_term = 0.5 * self.El * (phi_array + self.phase)**2
            junction_term = -self.Delta * np.sqrt(1 - self.T * np.sin(phi_array/2)**2)
        elif self.flux_grouping == 'ABS':
            inductive_term = 0.5 * self.El * phi_array**2
            junction_term = -self.Delta * np.sqrt(1 - self.T * np.sin((phi_array - self.phase)/2)**2)
            
        return inductive_term + junction_term
        
    def wavefunction(
        self,
        which: int = 0, 
        phi_grid: np.ndarray = None, 
        esys: Tuple[np.ndarray, np.ndarray] = None
        ) -> Dict[str, Any]:
        """
        Returns a wave function in the phi basis.

        Parameters
        ----------
        which : int, optional
            Index of desired wave function (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.

        Returns
        -------
        Dict[str, Any]
            Wave function data containing basis labels, amplitudes, and energy.
        """
        if esys is None:
            evals_count = max(which + 1, 3)
            evals, evecs = self.eigensys(evals_count)
        else:
            evals, evecs = esys
            
        dim = self.dimension
                        
        if phi_grid is None:
            phi_grid = np.linspace(-5 * np.pi, 5 * np.pi, 151)

        phi_basis_labels = phi_grid
        wavefunc_osc_basis_amplitudes = evecs[:, which]
        phi_wavefunc_amplitudes = np.zeros_like(phi_grid, dtype=np.complex128)
        
        for n in range(dim):
            phi_wavefunc_amplitudes += wavefunc_osc_basis_amplitudes[n] * self.harm_osc_wavefunction(n, phi_basis_labels, self.phase_zpf)

        return {
            "basis_labels": phi_basis_labels,
            "amplitudes": phi_wavefunc_amplitudes,
            "energy": evals[which]
        }

    def plot_wavefunction(
        self, 
        which: Union[int, Iterable[int]] = 0, 
        phi_grid: np.ndarray = None, 
        esys: Tuple[np.ndarray, np.ndarray] = None, 
        scaling: Optional[float] = None,
        **kwargs
        ) -> Tuple[plt.Figure, plt.Axes]:
        """
        Plot the wave function in the phi basis.

        Parameters
        ----------
        which : Union[int, Iterable[int]], optional
            Index or indices of desired wave function(s) (default is 0).
        phi_grid : np.ndarray, optional
            Custom grid for phi; if None, a default grid is used.
        esys : Tuple[np.ndarray, np.ndarray], optional
            Precomputed eigenvalues and eigenvectors.
        **kwargs
            Additional arguments for plotting. Can include:
            - fig_ax: Tuple[plt.Figure, plt.Axes], optional
                Figure and axes to use for plotting. If not provided, a new figure and axes are created.

        Returns
        -------
        Tuple[plt.Figure, plt.Axes]
            The figure and axes of the plot.
        """
        if isinstance(which, int):
            which = [which]
            
        potential = self.potential(phi=phi_grid)

        fig_ax = kwargs.get("fig_ax")
        if fig_ax is None:
            fig, ax = plt.subplots()
            fig.suptitle(self._generate_suptitle())
        else:
            fig, ax = fig_ax
        
        ax.plot(phi_grid/2/np.pi, potential, color='black', label='Potential')

        for idx in which:
            wavefunc_data = self.wavefunction(which=idx, phi_grid=phi_grid, esys=esys)
            phi_basis_labels = wavefunc_data["basis_labels"]
            wavefunc_amplitudes = wavefunc_data["amplitudes"]
            wavefunc_energy = wavefunc_data["energy"]

            ax.plot(
                phi_basis_labels/2/np.pi,
                wavefunc_energy + scaling * (wavefunc_amplitudes.real + wavefunc_amplitudes.imag),
                # color="blue",
                label=rf"$\Psi_{idx}$"
                )

        ax.set_xlabel(r"$\Phi / \Phi_0$")
        ax.set_ylabel(r"$\psi(\varphi)$, Energy [GHz]")
        ax.legend()
        ax.grid(True)

        return fig, ax
        

        


import os
import h5py
import numpy as np
from qutip import Qobj
from typing import Any, Dict, List, Union, Tuple

class SpectrumData:
    def __init__(
        self,
        energy_table: np.ndarray,
        system_params: Dict[str, Any],
        param_name: str = None,
        param_vals: np.ndarray = None,
        state_table: Union[List[Qobj], np.ndarray] = None,
        matrixelem_table: Dict[str, np.ndarray] = None,
        t1_table: Dict[str, np.ndarray] = None,
        tphi_table: Dict[Tuple[int, int, str], np.ndarray] = None,
        **kwargs
    ) -> None:
        
        self.system_params = system_params
        self.param_name = param_name
        self.param_vals = param_vals
        self.energy_table = energy_table
        self.state_table = state_table if state_table is not None else []
        self.matrixelem_table = matrixelem_table if matrixelem_table is not None else {}
        self.t1_table = t1_table if t1_table is not None else {}
        self.tphi_table = tphi_table if tphi_table is not None else {}
        for dataname, data in kwargs.items():
            setattr(self, dataname, data)

    def subtract_ground(self) -> None:
        """Subtract ground state energies from spectrum"""
        self.energy_table -= self.energy_table[:, 0]

    def filewrite(self, filename: str, overwrite: bool = False) -> None:
        """Save the SpectrumData to an HDF5 file."""
        
        if not overwrite and os.path.exists(filename):
            raise FileExistsError(f"The file '{filename}' already exists and overwrite is set to False.")
        
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with h5py.File(filename, 'w') as f:
            f.create_dataset('energy_table', data=self.energy_table)
            f.create_dataset('param_vals', data=self.param_vals)
            f.create_dataset('param_name', data=np.string_(self.param_name))
            f.create_dataset('system_params', data=np.string_(str(self.system_params)))
            
            if isinstance(self.state_table, list) and isinstance(self.state_table[0], Qobj):
                state_data = np.array([state.full() for state in self.state_table])
                f.create_dataset('state_table', data=state_data)
            else:
                f.create_dataset('state_table', data=self.state_table)
                
            if self.matrixelem_table:
                for key, value in self.matrixelem_table.items():
                    f.create_dataset(f'matrixelem_table/{key}', data=value)
            if self.t1_table:
                for key, value in self.t1_table.items():
                    f.create_dataset(f't1_table/{key}', data=value)
            if self.tphi_table:
                for key, value in self.tphi_table.items():
                    f.create_dataset(f'tphi_table/{key}', data=value)

    @staticmethod
    def read(filename: str) -> 'SpectrumData':
        """Read SpectrumData from an HDF5 file."""
        with h5py.File(filename, 'r') as f:
            energy_table = f['energy_table'][:]
            param_vals = f['param_vals'][:]
            param_name = f['param_name'][()].decode('utf-8')
            system_params = eval(f['system_params'][()].decode('utf-8'))
            state_table = f['state_table'][:] if 'state_table' in f else None
            matrixelem_table = {key: f[f'matrixelem_table/{key}'][:] for key in f['matrixelem_table']} if 'matrixelem_table' in f else {}
            t1_table = {tuple(key.split(',')): f[f't1_table/{key}'][:] for key in f['t1_table']} if 't1_table' in f else {}
            tphi_table = {tuple(key.split(',')): f[f'tphi_table/{key}'][:] for key in f['tphi_table']} if 'tphi_table' in f else {}
        
        return SpectrumData(
            energy_table=energy_table,
            system_params=system_params,
            param_name=param_name,
            param_vals=param_vals,
            state_table=state_table,
            matrixelem_table=matrixelem_table,
            t1_table=t1_table,
            tphi_table=tphi_table
        )